// 
// Copyright (c) 2006-2010, Benjamin Kaufmann
// 
// This file is part of Clasp. See http://www.cs.uni-potsdam.de/clasp/ 
// 
// Clasp is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// Clasp is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Clasp; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
//
#include "clasp_app.h"
#include "alarm.h"
#include <iostream>
#include <fstream>
#include <clasp/smodels_constraints.h>
#include <clasp/clause.h>
/////////////////////////////////////////////////////////////////////////////////////////
// Application
/////////////////////////////////////////////////////////////////////////////////////////
namespace Clasp {
#if !defined(CLASP_USAGE)
#define CLASP_USAGE   "clasp [number] [options]"
#endif
namespace {
inline std::string prettify(const std::string& str) {
	if (str.size() < 40) return str;
	std::string t("...");
	t.append(str.end()-38, str.end());
	return t;
}
std::ostream& operator << (std::ostream& os, Literal l) {
	if (l.sign()) os << '-';
	os << l.var();
	return os;
}
std::istream& operator >> (std::istream& in, Literal& l) {
	int i;
	if (in >> i) {
		l = Literal(i >= 0 ? Var(i) : Var(-i), i < 0);
	}
	return in;
}

}
/////////////////////////////////////////////////////////////////////////////////////////
// public functions & basic helpers
/////////////////////////////////////////////////////////////////////////////////////////
Application::Application() : timeToFirst_(-1.0), timeToLast_(-1.0), facade_(0), stopped_(0)  {}
Application& Application::instance() {
	static Application inst;
	return inst;
}
void Application::sigHandler(int sig) {
	Application::instance().kill(sig == SIGALRM ? reason_timeout : reason_interrupt);
}

// Kills any pending alarm
void Application::killAlarm() {
	if (cmdOpts_.basic.timeout>0) {
		setAlarm(0); 
	}
}

// Called on timeout or signal.
// Prints summary and then brutally kills the application.
void Application::kill(ReasonEnd r) {
	if (stopped_ == 0) {
		++stopped_; // ignore further kill requests
		SCOPE_ALARM_LOCK();
		for (int i = 0; i != sizeof(timer_)/sizeof(Timer); ++i) { 
			timer_[i].stop(); 
		}
		fprintf(stderr, "\n*** INTERRUPTED! ***\n");
		if (facade_ && facade_->state() != ClaspFacade::state_not_started) {
			printResult(r);
		}
		_exit(solver_.stats.solve.models != 0 ?  S_SATISFIABLE : S_UNKNOWN);
	}
}

void Application::installSigHandlers() {
	if (signal(SIGINT, &Application::sigHandler) == SIG_IGN) {
		signal(SIGINT, SIG_IGN);
	}
	if (signal(SIGTERM, &Application::sigHandler) == SIG_IGN) {
		signal(SIGTERM, SIG_IGN);
	}
	if (cmdOpts_.basic.timeout > 0) {
		setAlarmHandler(&Application::sigHandler);
		if (setAlarm(cmdOpts_.basic.timeout) == 0) {
			messages.warning.push_back("Could not set time limit!");
		}
	}
}

std::istream& Application::getStream() {
	if (generic.input.empty() || generic.input[0] == "stdin") {
		generic.input.resize(1, "stdin");
		return std::cin;
	}
	else {
		static std::ifstream file;
		if (file.is_open()) return file;
		file.open(generic.input[0].c_str());
		if (!file) { throw std::runtime_error("Can not read from '"+generic.input[0]+"'");  }
		return file;
	}
}

void Application::printHelp() const {
	printf("clasp %s\n", CLASP_VERSION);
	printf("\nusage: %s\n", CLASP_USAGE);
	printf("%s\n", getHelp().c_str());
	printf("\nusage: %s\n", CLASP_USAGE);
	printf("Default commandline:\nclasp %s\n", getDefaults().c_str());
	fflush(stdout);
}

void Application::printVersion() const {
	printf("clasp %s\n", CLASP_VERSION);	
	printf("Configured with: WITH_CLASPRE=%d\n", WITH_CLASPRE);
	printf("%s\n", CLASP_LEGAL);
	fflush(stdout);
}

bool Application::reportAndExit() const {
	if (generic.help) {
		printHelp();
		return true;
	}
	if (generic.version) {
		printVersion();
		return true;
	}
	if (claspre_.listFeatures) {
		Claspre::printFeatures();
		return true;
	}
	return false;
}

OutputFormat* Application::createOutput(Input::Format f) {
	if (claspre_.features != 0) {
		if (f == Input::SMODELS) {
			if (!config_.enumerate.limits.get()) {
				config_.enumerate.limits.reset(new std::pair<int, int>(500,20));
			}
			return Claspre::createOutput(solver_, static_cast<Claspre::Options::FeatureFormat>(claspre_.features), cmdOpts_, generic.verbose);
		}
		else {
			warning("Feature extraction not supported for current input format!");
		}
	}
	if (f == Input::SMODELS) {
		if (config_.onlyPre)     { generic.verbose = 0; }
		if (cmdOpts_.basic.asp09){ generic.verbose = 0; }
		return new AspOutput(cmdOpts_.basic.asp09);
	}
	else if (f == Input::DIMACS) { return new SatOutput(); }
	else if (f == Input::OPB)    { return new PbOutput(generic.verbose > 1);  }
	return 0;
}
/////////////////////////////////////////////////////////////////////////////////////////
// run - clasp's "main"-function
/////////////////////////////////////////////////////////////////////////////////////////
int Application::run(int argc, char** argv) {
	if (!parse(argc, argv, parsePositional)) {
		// command-line error
		throw std::runtime_error( messages.error.c_str() );
	}
	if (reportAndExit()) {
		return EXIT_SUCCESS;
	}
	facade_ = 0;
	installSigHandlers();
	StreamInput input(getStream(), detectFormat(getStream()));
	cmdOpts_.applyDefaults(input.format());
	out_.reset(createOutput(input.format()));
	if (generic.verbose >= 2 && solver_.strategies().satPrePro.get()) {
		solver_.strategies().satPrePro->setVerbose(&Application::preproProgress);
	}
	if (cmdOpts_.basic.stats > 1) {
		solver_.stats.solve.enableJumpStats();
	}
	ClaspFacade clasp; facade_ = &clasp;
	timer_[0].start();
	clasp.solve(input, config_, this);
	timer_[0].stop();
	++stopped_;  // disable signal handler
	killAlarm(); // and kill any pending alarms;
	printResult(reason_end);
	if      (clasp.result() == ClaspFacade::result_unsat) return S_UNSATISFIABLE;
	else if (clasp.result() == ClaspFacade::result_sat)   return S_SATISFIABLE;
	else                                                  return S_UNKNOWN;
}

/////////////////////////////////////////////////////////////////////////////////////////
// State & Result functions
/////////////////////////////////////////////////////////////////////////////////////////
// Prints a summary after search has stopped or has been interrupted.
// The summary prints a solution line (defined by the output format)
// If --verbose > 0 the summary furthermore consists of:
//  - the number of computed models m followed by a + if search was not completed
//  - the number of enumerated models e if e != m
//  - the state of any optimization and whether the last model was optimal
//  - the state of consequence computation and whether the last model corresponded to the consequences
//  - the time spent
// Finally, if --stats is given, output format defined statistics are printed.
void Application::printResult(ReasonEnd end) const {
	bool complete        = end == reason_end && !facade_->more();	
	const Solver& s      = solver_;
	const Enumerator& en = *config_.solve.enumerator();
	if (config_.onlyPre) {
		if (end != reason_end) return;
		if (facade_->api()) { // asp-mode
			facade_->result() == ClaspFacade::result_unsat
				? (void)(std::cout << "0\n0\nB+\n1\n0\nB-\n1\n0\n0\n")
				: facade_->api()->writeProgram(std::cout);
			delete facade_->releaseApi();
		}
		else {
			if (facade_->result() != ClaspFacade::result_unsat) {
				status(0, "Search not started because of option '--pre'!\n");
			}
			out_->printSolution(solver_, en, complete);
		}
		return;
	}
	out_->printSolution(s, en, complete);
	if (cmdOpts_.basic.quiet && config_.enumerate.consequences() && s.stats.solve.models != 0) {
		status(1, "%s consequences:\n", config_.enumerate.cbType());
		out_->printConsequences(s, en);
	}
	if (generic.verbose > 0) {
		const char* c= out_->format[OutputFormat::comment];
		const int   w= 12-(int)strlen(c);
		if      (end == reason_timeout)  { fprintf(STATUS_CHANNEL, "\n%sTIME LIMIT  : 1\n", c); }
		else if (end == reason_interrupt){ fprintf(STATUS_CHANNEL, "\n%sINTERRUPTED : 1\n", c); }
		uint64 enumerated = s.stats.solve.models;
		uint64 models     = enumerated;
		if      (config_.enumerate.consequences() && enumerated > 1) { models = 1; }
		else if (en.minimize())                                { models = en.minimize()->models(); }
		fprintf(STATUS_CHANNEL, "\n%s%-*s: ", c, w, "Models");
		if (!complete) {
			char buf[64];
			int wr    = sprintf(buf, "%"PRIu64, models);
			buf[wr]   = '+';
			buf[wr+1] = 0;
			fprintf(STATUS_CHANNEL, "%-6s\n", buf);
		}
		else {
			fprintf(STATUS_CHANNEL, "%-6"PRIu64"\n", models);
		}
		if (enumerated) {
			if (enumerated != models) { 
				fprintf(STATUS_CHANNEL, "%s%-*.*s: %"PRIu64"\n", c, w, w, "  Enumerated", enumerated);
			}
			if (config_.enumerate.consequences()) {
				fprintf(STATUS_CHANNEL, "%s  %-*s: %s\n", c, w-2, config_.enumerate.cbType(), complete?"yes":"unknown");
			}
			if (en.minimize()) {
				fprintf(STATUS_CHANNEL, "%s%-*s: %s\n", c, w, "  Optimum", complete?"yes":"unknown");
				fprintf(STATUS_CHANNEL, "%s%-*s: ", c, w, "Optimization");
				out_->printOptimizeValues(*en.minimize());
				fprintf(STATUS_CHANNEL, "\n");
			}
		}
		double solve = facade_->state() == ClaspFacade::state_solve ? timer_[ClaspFacade::state_solve].total() : 0.0;
		double ttl   = timeToLast_ != -1.0 ? timeToLast_ : 0.0;
		fprintf(STATUS_CHANNEL, "%s%-*s: %.3fs (Solving: %.2fs 1st Model: %.2fs Unsat: %.2fs)\n", c, w, "Time", timer_[0].total()
			, solve
			, timeToFirst_ != -1.0 ? timeToFirst_ : 0.0
			, complete && solve-ttl >= 0.001 ? solve-ttl : 0.0);
	}
	if (cmdOpts_.basic.stats) { out_->printStats(s.stats, en); }
	if (!lemmaOut.empty()) {
		try { writeLemmas(); } 
		catch (const std::exception& e) {
			ERROR_OUT("clasp ERROR: ", e.what());
		}
	}
}

// State-transition callback.
// Updates status information and handles timing
void Application::state(ClaspFacade::Event e, ClaspFacade& f) { 
	SCOPE_ALARM_LOCK();
	if (e == ClaspFacade::event_state_enter) {
		while (!messages.warning.empty()) {
			fprintf(stderr, "Warning: %s\n", messages.warning.back().c_str());
			messages.warning.pop_back();
		}
		if (f.state() == ClaspFacade::state_read) {
			status(1, "clasp version %s\n", CLASP_VERSION);
			status(1, "Reading from %s\n", prettify(generic.input[0]).c_str());
			status(2, "Reading      : ");
		}	
		else if (f.state() == ClaspFacade::state_preprocess) {
			status(2, "Preprocessing: ");
		}
		else if (f.state() == ClaspFacade::state_solve) {
			status(1, "Solving...\n");
		}
		timer_[f.state()].start();
	}
	else if (e == ClaspFacade::event_state_exit) {
		timer_[f.state()].stop();
		if (generic.verbose > 1 && (f.state() == ClaspFacade::state_read || f.state() == ClaspFacade::state_preprocess)) {
			fprintf(STATUS_CHANNEL, "%.3f\n", timer_[f.state()].total());
		}
	}
}

// Event callback.
void Application::event(ClaspFacade::Event e, ClaspFacade& f) {
	if (e == ClaspFacade::event_model) {
		timer_[f.state()].lap();
		timeToLast_ = timer_[f.state()].total();
		if (timeToFirst_ == -1.0) {  timeToFirst_ = timeToLast_; }
		if (!cmdOpts_.basic.quiet) { 
			SCOPE_ALARM_LOCK();
			if ( !(config_.enumerate.consequences()) ) {
				status(1, "Answer: %"PRIu64"\n", solver_.stats.solve.models);
				out_->printModel(solver_, *config_.solve.enumerator());
			}
			else {
				status(1, "%s consequences:\n", config_.enumerate.cbType());
				out_->printConsequences(solver_, *config_.solve.enumerator());
			}
			if (config_.solve.enumerator()->minimize()) {
				out_->printOptimize(*config_.solve.enumerator()->minimize());
			}
		}
	}
	else if (e == ClaspFacade::event_p_prepared) {
		if (config_.onlyPre) {
			if (f.api()) f.releaseApi(); // keep api so that we can later print the program
			else         status(0, "Vars: %u, Constraints: %u\n", solver_.numVars(), solver_.numConstraints());
			AtomIndex* x = solver_.strategies().symTab.release();
			solver_.reset(); // release constraints and strategies - no longer needed
			solver_.strategies().symTab.reset(x);
		}
		else { 
			if (!lemmaIn.empty()) {
				readLemmas();	
			}
			if (generic.verbose >= 2) {
				f.config()->solve.enumerator()->enableProgressReport(&f);
			}
			out_->initSolve(solver_, f.api(), f.config()->solve.enumerator());
		}
	}
}

void Application::reportRestart(const Solver& s, uint64 maxC, uint32 maxL) {
	status(2, "V: %7u, C: %8u, L: %8u, ML: %8u, R: %8"PRIu64"\n"
		, s.numFreeVars()
		, s.numConstraints()
		, s.numLearntConstraints()
		, maxL
		, maxC);
}
/////////////////////////////////////////////////////////////////////////////////////////
// Status messages
/////////////////////////////////////////////////////////////////////////////////////////
// If --verbose >= v, prints (%s%s,format[comment],MSG) to STATUS_CHANNEL,
// where MSG is the message given in the parameters.
void Application::status(int v, const char* fmt, ...) const {
	if (generic.verbose >= v) {
		fprintf(STATUS_CHANNEL, "%s", out_->format[OutputFormat::comment]);
		va_list args;
		va_start(args, fmt);
		vfprintf(STATUS_CHANNEL, fmt, args);
		va_end(args);
	}
}	
// Prints progress of Sat-Preprocessor.
// Only used if --verbose >= 2.
void Application::preproProgress(SatPreprocessor::State a, uint32 i) {
	static const char pro[] = {'/', '-', '\\', '|' };
	if      (a==SatPreprocessor::pre_start)   fprintf(STATUS_CHANNEL, "|");
	else if (a==SatPreprocessor::iter_start)  fprintf(STATUS_CHANNEL, "\b%c", pro[i%4]);
	else if (a==SatPreprocessor::pre_done)    fprintf(STATUS_CHANNEL, "\b");
	else if (a==SatPreprocessor::pre_stopped) fprintf(STATUS_CHANNEL, "\b");
	fflush(STATUS_CHANNEL);
}

/////////////////////////////////////////////////////////////////////////////////////////
// internal helpers
/////////////////////////////////////////////////////////////////////////////////////////
// NOTE: ON WINDOWS this function is unsafe if called from time-out handler becaue
// it has potential races with the main thread
void Application::writeLemmas() const {
	std::ofstream file(lemmaOut.c_str());
	if (!file) throw std::runtime_error("Could not open file: "+lemmaOut);
	for (LitVec::size_type i = 0; i != solver_.numLearntConstraints(); ++i) {
		if (solver_.getLearnt(i).type() == Constraint_t::learnt_conflict) {
			const Clause& c = (const Clause&)solver_.getLearnt(i);
			for (uint32 x = 0; x != c.size(); ++x) {
				file << c[x] << " ";
			}
			if (!isSentinel(*c.end())) {
				const Literal* tail = c.end();
				file << ~(*tail++) << " ";
				while (!isSentinel(*tail)) {
					file << (*tail++) << " ";
				}
			}
			file << "0\n";
			if (!file) throw std::runtime_error("Error writing file: "+lemmaOut);
		}
	}
}

void Application::readLemmas() {
	std::ifstream file(lemmaIn.c_str());
	if (!file) throw std::runtime_error("Could not open file: "+lemmaIn);
	ClauseCreator clause(&solver_);
	for (;file;) {
		for (std::string ignore; file && (file.peek() == 'c' || file.peek() == 'p'); ) {
			std::getline(file, ignore);
		}
		clause.start();
		Literal x;
		for (; file >> x && x.var() != 0; ) { 
			if (!solver_.validVar(x.var())) throw std::runtime_error("Bad variable in file: "+lemmaIn);
			clause.add(x);
		}
		if (!file || !clause.end()) {
			break;
		}
	}
	if (!file.eof() && !solver_.hasConflict()) {
		throw std::runtime_error("Error reading file: "+lemmaIn);
	}
}

} // end of namespace clasp

